"""Examples for the repo README.

The README has a "Quick Examples" section.
This is a tough place for examples:

- It's outside of Sphinx, to be served by GitHub
- But it's included into the top of the Sphinx docs

Because it isn't using Sphinx, we can't do includes.
Thus, we have to do cut-and-paste, so let's keep it to a minimum.
"""
